// Cool Edit filter for FLAC - Free Lossless Audio Codec
// http://flac.sourceforge.net
// Copyright (C) 2003 Janne Hyvrinen
//
// Changes:
//  0.1.8 (2003-07-27): Fixed save failure on mono files
//  0.1.7 (2003-07-05): Fixed divide by zero crash on very short files
//  0.1.6 (2003-05-25): Fixed crashing on filetype detection
//  0.1.5 (2003-05-25): Major code cleanup, option dialog, support for OggFLAC
//  0.1.4 (2003-05-17): Removed unnecessary seektable
//  0.1.3 (2003-05-15): Fixed file saving
//  0.1.2 (2003-04-04): Proper file type detection, no longer plain extension check,
//                      Fixed bug: file handle wasn't closed when writing finished,
//                      Fixed bug: other than 16 bits files didn't work
//  0.1.1 (2003-04-03): One function needed by writing was missing, small improvements
//  0.1.0 (2003-04-02): Initial version, supports reading

#include <windows.h>
#include <commctrl.h>
#include "res/filters.h"
#include "res/resource.h"
#include "flac_lib.h"

#define CF_ERRORMESSAGES    // MessageBox on errors
//#define CF_ALL_TO_32BIT     // Convert everything to 32bit float

#define CF_MAX_CHANNELS     2
#define CF_SAMPLES          1152
#define CHUNKSIZE           FLAC__MAX_BLOCK_SIZE * CF_MAX_CHANNELS * 32/8

// -------------------------------------

typedef struct {
    flac_decoder *decoder;
    Reader *r;

    FLAC__byte sample_buffer[CHUNKSIZE];
    int bpos, bwaiting;

    unsigned decode_bps;
    DWORD Options;
} INPUT;

typedef struct {
    flac_encoder *encoder;
    Reader *r;

    unsigned channels, source_bps, oggflac;
    DWORD Options;
} OUTPUT;


void infobox ( const char *msg )
{
# ifdef CF_ERRORMESSAGES
    MessageBox ( 0, msg, "FLAC", 0 );
# endif
}

// Options ------

__declspec (dllexport) DWORD FAR PASCAL FilterGetOptions ( HWND hWnd, HINSTANCE hInst, long lSamprate, WORD wChannels, WORD wBitsPerSample, DWORD dwOptions )
{
    long nDialogReturn = 0;
    FARPROC lpfnDIALOGMsgProc;
    lpfnDIALOGMsgProc = GetProcAddress ( hInst, (LPCSTR)MAKELONG(20, 0) );

    if ( dwOptions == 0 )   // change default to 5
        nDialogReturn = 5;
    else
        nDialogReturn = dwOptions;

    nDialogReturn = (long)DialogBoxParam ( (HINSTANCE)hInst, (LPCSTR)MAKEINTRESOURCE(IDD_CONFIG), (HWND)hWnd, (DLGPROC)lpfnDIALOGMsgProc, nDialogReturn&0xFF );

    return nDialogReturn | (1<<15);
}

__declspec (dllexport) DWORD FAR PASCAL FilterOptions ( HANDLE hInput )
{
    if ( !hInput ) return 0;

    DWORD options = 0;
    INPUT far *i = (INPUT far *)GlobalLock ( hInput );
    options = i->Options;
    GlobalUnlock ( hInput );

    return options;
}

__declspec (dllexport) DWORD FAR PASCAL FilterOptionsString ( HANDLE hInput, LPSTR szString )
{
    if ( !hInput || szString == NULL ) return 0;

    DWORD options = 0;
    INPUT far *i = (INPUT far *)GlobalLock ( hInput );

    if ( i->decoder ) {
        double size = (double)i->decoder->pReader->get_length();
        double dur = (i->decoder->streaminfo.data.stream_info.sample_rate > 0)  ?  (double)(__int64)i->decoder->streaminfo.data.stream_info.total_samples / (double)i->decoder->streaminfo.data.stream_info.sample_rate  :  0;
        double bitrate = (dur > 0)  ?  (size * 8.) / (dur * 1000.)  :  0;

        wsprintf ( szString, "FLAC\nBitrate: %u kbps", (unsigned int)bitrate );
        options = i->Options;
    }

    GlobalUnlock ( hInput );

    return options;
}

__declspec (dllexport) BOOL FAR PASCAL DIALOGMsgProc ( HWND hWndDlg, UINT Message, WPARAM wParam, LPARAM lParam )
{
    char str[128];
    HWND hWndCSlider = GetDlgItem ( hWndDlg, IDC_COMPRESSION );
    HWND hWndCValue  = GetDlgItem ( hWndDlg, IDC_COMPRESSION_VALUE );

    switch ( Message ) {
    case WM_INITDIALOG:
        {
            long compression = (long)lParam & 0xFF;
            long oggflac = ((long)lParam & (1<<8)) != 0;
            if ( compression < 0 || compression > 8 ) compression = 5;

            SendDlgItemMessage ( hWndDlg, IDC_COMPRESSION, TBM_SETRANGEMIN, 0, 0 );
            SendDlgItemMessage ( hWndDlg, IDC_COMPRESSION, TBM_SETRANGEMAX, 0, 8 );
            SendDlgItemMessage ( hWndDlg, IDC_COMPRESSION, TBM_SETPOS, 1, compression );

            for ( int i = 0; i <= 8; i++ ) {
                SendDlgItemMessage ( hWndDlg, IDC_COMPRESSION, TBM_SETTIC, 1, i );
            }

            wsprintf ( str, "%u", compression );
            SetDlgItemText ( hWndDlg, IDC_COMPRESSION_VALUE, str );

            CheckDlgButton ( hWndDlg, IDC_OGGFLAC, oggflac );

            UpdateWindow ( hWndDlg );
        }
        break;

    case WM_HSCROLL:
        if ( (HWND)lParam == hWndCSlider ) {
            long compression = SendDlgItemMessage ( hWndDlg, IDC_COMPRESSION, TBM_GETPOS, 0, 0 );
            wsprintf ( str, "%u", compression );
            SetDlgItemText ( hWndDlg, IDC_COMPRESSION_VALUE, str );
            UpdateWindow ( hWndDlg );
            return TRUE;
        }
        break;

    case WM_COMMAND:
        switch ( LOWORD (wParam) ) {
        case IDOK:
            {
                long compression = SendDlgItemMessage ( hWndDlg, IDC_COMPRESSION, TBM_GETPOS, 0, 0 );
                long oggflac = IsDlgButtonChecked ( hWndDlg, IDC_OGGFLAC );
                EndDialog ( hWndDlg, (short)(compression | (oggflac << 8) | (1<<15)) );
            }
            break;

        case IDCANCEL:
             EndDialog ( hWndDlg, FALSE );
             break;
        }
        break;

    case WM_CLOSE:
        PostMessage ( hWndDlg, WM_COMMAND, IDCANCEL, 0L );
        break;

    default:
        return FALSE;
    }

    return TRUE;
}

// End Options ---
// General ------

__declspec (dllexport) short FAR PASCAL QueryCoolFilter ( COOLQUERY far *cq )
{
    strcpy ( cq->szName,      "FLAC" );
    strcpy ( cq->szCopyright, "Free Lossless Audio Codec" );
    strcpy ( cq->szExt,       "FLA" );
    strcpy ( cq->szExt2,      "FL*" );
    strcpy ( cq->szExt3,      "OGG" );
    strcpy ( cq->szExt4,      "" );

    cq->lChunkSize = CHUNKSIZE;
    cq->dwFlags    = QF_CANLOAD | QF_CANSAVE | QF_HASOPTIONSBOX | QF_RATEADJUSTABLE | QF_CANDO32BITFLOATS;
    cq->Stereo8    = 0xFF;
    cq->Stereo16   = 0xFF;
    cq->Stereo24   = 0xFF;
    cq->Stereo32   = 0xFF;
    cq->Mono8      = 0xFF;
    cq->Mono16     = 0xFF;
    cq->Mono24     = 0xFF;
    cq->Mono32     = 0xFF;
    cq->Quad32     = 0xFF;
    cq->Quad16     = 0xFF;
    cq->Quad8      = 0xFF;

    return C_VALIDLIBRARY;
}

__declspec (dllexport) BOOL FAR PASCAL FilterUnderstandsFormat ( LPSTR lpstrFilename )
{
    if ( !lpstrFilename || !lpstrFilename[0] ) return FALSE;

    char *ext = strrchr ( lpstrFilename, '.' );
    if ( !ext ) return FALSE;
    if ( strnicmp (ext, ".fl", 2) && stricmp (ext, ".ogg") ) return FALSE;

    HANDLE fh = CreateFile ( lpstrFilename, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );
    if ( fh == INVALID_HANDLE_VALUE ) return FALSE;

    Reader *tr = new Reader_file ( fh );
    if ( !tr ) {
        CloseHandle ( fh );
        return FALSE;
    }

    BOOL supported = flac_decoder::is_flac ( tr );
    if ( !supported ) supported = flac_decoder::is_ogg_flac ( tr );

    delete tr;

    return supported;
}

__declspec (dllexport) VOID FAR PASCAL GetSuggestedSampleType ( LONG *lplSamprate, WORD *lplwBitsPerSample, WORD *lpwChannels )
{
    int sr  = *lplSamprate;
    int ch  = *lpwChannels;
    int bps = *lplwBitsPerSample;

    *lplSamprate = 0;
    *lpwChannels = 0;
    *lplwBitsPerSample = 0;

    if ( bps < FLAC__MIN_BITS_PER_SAMPLE ) *lplwBitsPerSample = FLAC__MIN_BITS_PER_SAMPLE;
    else if ( bps > FLAC__MIN_BITS_PER_SAMPLE && bps < 8 ) *lplwBitsPerSample = 8;
    else if ( bps > 8 && bps < 16 ) *lplwBitsPerSample = 16;
    else if ( bps > 16 && bps < 24 ) *lplwBitsPerSample = 24;

    if ( sr  > FLAC__MAX_SAMPLE_RATE ) *lplSamprate = FLAC__MAX_SAMPLE_RATE;
    if ( ch  > CF_MAX_CHANNELS ) *lpwChannels = CF_MAX_CHANNELS;
    if ( bps > FLAC__REFERENCE_CODEC_MAX_BITS_PER_SAMPLE ) *lplwBitsPerSample = FLAC__REFERENCE_CODEC_MAX_BITS_PER_SAMPLE;
}

__declspec (dllexport) long FAR PASCAL FilterGetFileSize ( HANDLE hInput )
{
    long size_in_bytes = 0;

    if ( hInput ) {
        INPUT FAR *i = (INPUT FAR *)GlobalLock ( hInput );
        if ( i->decoder ) size_in_bytes = (long)(i->decoder->streaminfo.data.stream_info.total_samples * i->decoder->streaminfo.data.stream_info.channels * i->decode_bps/8);
        GlobalUnlock ( hInput );
    }

    return size_in_bytes;
}

// End General ---
// Reading ------

__declspec (dllexport) HANDLE FAR PASCAL OpenFilterInput ( LPSTR lpstrFilename, long far *lSamprate, WORD far *wBitsPerSample, WORD far *wChannels, HWND hWnd, long far *lChunkSize )
{
    HANDLE hInput;

    hInput = GlobalAlloc ( GMEM_MOVEABLE | GMEM_SHARE | GMEM_ZEROINIT, sizeof (INPUT) );
    if ( hInput ) {
        INPUT far *i = (INPUT far *)GlobalLock ( hInput );

        HANDLE fh = CreateFile ( lpstrFilename, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN, NULL );
        if ( fh == INVALID_HANDLE_VALUE ) {
            GlobalUnlock ( hInput );
            GlobalFree ( hInput );
            return NULL;
        }

        i->r = new Reader_file ( fh );
        if ( !i->r ) {
            CloseHandle ( fh );
            GlobalUnlock ( hInput );
            GlobalFree ( hInput );
            return NULL;
        }

        i->decoder = new flac_decoder ( i->r );
        if ( !i->decoder ) {
            delete i->r;
            GlobalUnlock ( hInput );
            GlobalFree ( hInput );
            return NULL;
        }

        if ( !i->decoder->status_ok ) {
            delete i->decoder;
            delete i->r;
            GlobalUnlock ( hInput );
            GlobalFree ( hInput );
            return NULL;
        }

        if ( i->decoder->streaminfo.data.stream_info.channels > CF_MAX_CHANNELS ) {
            char temp[256];
            wsprintf ( temp, "Maximum of %i channels supported, file has %i.", CF_MAX_CHANNELS, (int)i->decoder->streaminfo.data.stream_info.channels );
            infobox ( temp );
            delete i->decoder;
            delete i->r;
            GlobalUnlock ( hInput );
            GlobalFree ( hInput );
            return NULL;
        }

        // Cool Edit only supports 8/16 bit PCM and 32 bit float
        i->decode_bps = i->decoder->streaminfo.data.stream_info.bits_per_sample;
        if ( i->decode_bps != 8 && i->decode_bps != 16 ) i->decode_bps = 32;

#ifdef CF_ALL_TO_32BIT
        i->decode_bps = 32;
#endif

        *lSamprate = i->decoder->streaminfo.data.stream_info.sample_rate;
        *wBitsPerSample = i->decode_bps;
        *wChannels = i->decoder->streaminfo.data.stream_info.channels;
        *lChunkSize = (*wChannels) * (*wBitsPerSample)/8 * CF_SAMPLES;

        GlobalUnlock ( hInput );
    }

    return hInput;
}

__declspec (dllexport) void FAR PASCAL CloseFilterInput ( HANDLE hInput )
{
    if ( hInput ) {
        INPUT far *i = (INPUT far *)GlobalLock ( hInput );

        if ( i->decoder ) delete i->decoder;
        if ( i->r ) delete i->r;

        GlobalUnlock ( hInput );
        GlobalFree ( hInput );
    }
}

__declspec (dllexport) DWORD FAR PASCAL ReadFilterInput ( HANDLE hInput, unsigned char far *buffer, long lBytes )
{
    long read = 0L;
    unsigned char far *buf = buffer;

    if (hInput) {
        INPUT far *i = (INPUT far *)GlobalLock ( hInput );
        long b;

        if ( i->bwaiting > 0 ) {
            b = ( i->bwaiting > lBytes )  ?  lBytes  :  i->bwaiting;
            memcpy ( buf, i->sample_buffer + i->bpos, b );
            buf += b;
            i->bwaiting -= b;
            if ( i->bwaiting == 0 ) i->bpos = 0; else i->bpos += b;
            read += b;
        }

        while ( read < lBytes ) {
            int samples = i->decoder->decode ( i->sample_buffer, CF_SAMPLES, (i->decode_bps == 32) );
            if ( samples <= 0 ) break;

            int bytes = samples * i->decoder->streaminfo.data.stream_info.channels * i->decode_bps/8;

            if ( (int)(read + bytes) <= (int)lBytes ) {
                b = bytes;
            } else {
                b = lBytes - read;
                i->bpos = b;
                i->bwaiting = bytes - i->bpos;
            }

            memcpy ( buf, i->sample_buffer, b );
            buf += b;
            read += b;
        }

        GlobalUnlock ( hInput );
    }

    return read;
}

// End Reading ---
// Writing ------

__declspec (dllexport) HANDLE FAR PASCAL OpenFilterOutput ( LPSTR lpstrFilename, long lSamprate, WORD wBitsPerSample, WORD wChannels, long lSize, long far *lpChunkSize, DWORD dwOptions )
{
    HANDLE hOutput;

    if ( !FLAC__format_sample_rate_is_valid (lSamprate) ) return NULL;

    hOutput = GlobalAlloc ( GMEM_MOVEABLE | GMEM_SHARE | GMEM_ZEROINIT, sizeof (OUTPUT) );
    if ( hOutput ) {
        OUTPUT far *o = (OUTPUT far *)GlobalLock ( hOutput );

        HANDLE fh = CreateFile ( lpstrFilename, GENERIC_READ|GENERIC_WRITE, FILE_SHARE_READ, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL );
        if ( fh == INVALID_HANDLE_VALUE ) {
            GlobalUnlock ( hOutput );
            GlobalFree ( hOutput );
            return NULL;
        }

        o->r = new Reader_file ( fh );
        if ( !o->r ) {
            CloseHandle ( fh );
            GlobalUnlock ( hOutput );
            GlobalFree ( hOutput );
            return NULL;
        }

        flac_encoder::flac_encoder_parameters parameters;

        o->channels = wChannels;
        o->source_bps = wBitsPerSample;
        o->Options = dwOptions;

        if ( dwOptions ) {
            flac_encoder::compression_level ( parameters, dwOptions & 0xFF );
            o->oggflac = (dwOptions & (1<<8)) != 0;
        } else {
            o->Options = 5 | (1<<15);
            o->oggflac = 0;
            flac_encoder::compression_level ( parameters, 5 );
        }

        parameters.channels = wChannels;
        parameters.bits_per_sample = wBitsPerSample;
        parameters.sample_rate = lSamprate;
        parameters.total_samples = lSize / (o->channels * o->source_bps/8);
        parameters.create_ogg = o->oggflac;
        parameters.verify = true;
        parameters.streamable_subset = false;

        // save float files as 24 bit PCM
        if ( wBitsPerSample > 24 ) parameters.bits_per_sample = 24;

        o->encoder = new flac_encoder ( o->r, parameters, 0 );
        if ( !o->encoder ) {
            delete o->r;
            GlobalUnlock ( hOutput );
            GlobalFree ( hOutput );
            infobox ( "Failed to create new encoder." );
            return NULL;
        }

        if ( !o->encoder->status_ok ) {
            infobox ( o->encoder->error_message );
            delete o->encoder;
            delete o->r;
            GlobalUnlock ( hOutput );
            GlobalFree ( hOutput );
            return NULL;
        }

        *lpChunkSize = CHUNKSIZE;

        GlobalUnlock ( hOutput );
    }

    return hOutput;
}

__declspec (dllexport) void FAR PASCAL CloseFilterOutput ( HANDLE hOutput )
{
    if ( hOutput ) {
        OUTPUT far *o = (OUTPUT far *)GlobalLock ( hOutput );

        if ( o->encoder ) delete o->encoder;
        if ( o->r ) delete o->r;

        GlobalUnlock ( hOutput );
        GlobalFree ( hOutput );
    }
}

__declspec (dllexport) DWORD FAR PASCAL WriteFilterOutput ( HANDLE hOutput, unsigned char far *buffer, long lBytes )
{
    if ( hOutput ) {
        OUTPUT far *o = (OUTPUT far *)GlobalLock ( hOutput );

        int samples = lBytes / (o->channels * o->source_bps/8);

        int _s = o->encoder->encode ( buffer, samples, o->source_bps );

        if ( _s != samples ) {
            infobox ( o->encoder->error_message );
            GlobalUnlock ( hOutput );
            return 0;
        }
    }

    GlobalUnlock ( hOutput );

    return lBytes;
}

// End Writing ---
// DllMain ------

BOOL WINAPI DllMain ( HANDLE hModule, DWORD fdwReason, LPVOID lpReserved )
{
   switch ( fdwReason ) {
      case DLL_PROCESS_ATTACH:
        /* Code from LibMain inserted here.  Return TRUE to keep the
            DLL loaded or return FALSE to fail loading the DLL.
 
            You may have to modify the code in your original LibMain to
            account for the fact that it may be called more than once.
            You will get one DLL_PROCESS_ATTACH for each process that
            loads the DLL. This is different from LibMain which gets
            called only once when the DLL is loaded. The only time this
            is critical is when you are using shared data sections.
            If you are using shared data sections for statically
            allocated data, you will need to be careful to initialize it
            only once. Check your code carefully.
 
            Certain one-time initializations may now need to be done for
            each process that attaches. You may also not need code from
            your original LibMain because the operating system may now
            be doing it for you.
         */
         break;
 
      case DLL_THREAD_ATTACH:
         /* Called each time a thread is created in a process that has
            already loaded (attached to) this DLL. Does not get called
            for each thread that exists in the process before it loaded
            the DLL.
 
            Do thread-specific initialization here.
         */
         break;
 
      case DLL_THREAD_DETACH:
         /* Same as above, but called when a thread in the process
            exits.
 
            Do thread-specific cleanup here.
         */
         break;
 
      case DLL_PROCESS_DETACH:
         /* Code from _WEP inserted here.  This code may (like the
            LibMain) not be necessary.  Check to make certain that the
            operating system is not doing it for you.
         */
         break;
   }
 
   /* The return value is only used for DLL_PROCESS_ATTACH; all other
      conditions are ignored.  */
   return TRUE;   // successful DLL_PROCESS_ATTACH
}

// End DllMain ---
