// FLAC helper library, see flac_lib.h for more info

#include "flac_lib.h"

// -------------------------------------

int flac_decoder::decode ( FLAC__byte *output, unsigned int samples, int decode_to_float=0 )
{
    if ( eof ) return 0;

    if ( seekable_stream_decoder ) {
        while ( samples_in_reservoir < samples ) {
            if ( FLAC__seekable_stream_decoder_get_state (seekable_stream_decoder) == FLAC__SEEKABLE_STREAM_DECODER_END_OF_STREAM ) break;
            if ( !FLAC__seekable_stream_decoder_process_single (seekable_stream_decoder) ) return 0;
        }
    }
    else if (stream_decoder) {
        while ( samples_in_reservoir < samples ) {
            unsigned _s = samples_in_reservoir;
            if ( FLAC__stream_decoder_get_state (stream_decoder) == FLAC__STREAM_DECODER_END_OF_STREAM ) break;
            if ( !FLAC__stream_decoder_process_single (stream_decoder) ) return 0;
            if ( _s == samples_in_reservoir ) break;
        }
    }
    else if (ogg_stream_decoder) {
        while ( samples_in_reservoir < samples ) {
            unsigned _s = samples_in_reservoir;
            if ( OggFLAC__stream_decoder_get_state (ogg_stream_decoder) == FLAC__SEEKABLE_STREAM_DECODER_END_OF_STREAM ) break;
            if ( !OggFLAC__stream_decoder_process_single (ogg_stream_decoder) ) return 0;
            if ( _s == samples_in_reservoir ) break;
        }
    }
    else return 0;

    const unsigned n = (samples_in_reservoir < samples)  ?  samples_in_reservoir  :  samples;
    const unsigned delta = n * streaminfo.data.stream_info.channels;
    unsigned j;

    int bytes;

    if ( !decode_to_float ) {
        bytes = pack_pcm ( output, reservoir, n, streaminfo.data.stream_info.channels, streaminfo.data.stream_info.bits_per_sample );
    } else {
        bytes = convert_to_float ( (float *)output, reservoir, n, streaminfo.data.stream_info.channels, streaminfo.data.stream_info.bits_per_sample );
    }

    for ( j = delta; j < samples_in_reservoir * streaminfo.data.stream_info.channels; j++ ) reservoir[j-delta] = reservoir[j];

    samples_in_reservoir -= n;

    return n;
}

flac_decoder::flac_decoder ( Reader *r )
{
    seekable_stream_decoder = NULL;
    stream_decoder = NULL;
    ogg_stream_decoder = NULL;

    samples_in_reservoir = 0;
    abort_flag = 0;
    eof = 0;

    reservoir = NULL;
    reservoir_size = 0;

    pReader = r;

    memset ( &streaminfo, 0, sizeof(streaminfo) );

    status_ok = false;

    if ( is_ogg_flac (pReader) ) { // OggFLAC file/stream
        ogg_stream_decoder = OggFLAC__stream_decoder_new();
        if ( !ogg_stream_decoder ) {
            return;
        }

        OggFLAC__stream_decoder_set_read_callback ( ogg_stream_decoder, readCallback_ogg_ns );
        OggFLAC__stream_decoder_set_write_callback ( ogg_stream_decoder, writeCallback_ogg_ns );
        OggFLAC__stream_decoder_set_metadata_callback ( ogg_stream_decoder, metadataCallback_ogg_ns );
        OggFLAC__stream_decoder_set_error_callback ( ogg_stream_decoder, errorCallback_ogg_ns );
        OggFLAC__stream_decoder_set_client_data ( ogg_stream_decoder, this );
        //OggFLAC__stream_decoder_set_metadata_respond ( ogg_stream_decoder, FLAC__METADATA_TYPE_VORBIS_COMMENT );

        if ( OggFLAC__stream_decoder_init (ogg_stream_decoder) != FLAC__SEEKABLE_STREAM_DECODER_OK ) {
            cleanup();
            return;
        }

        if ( !OggFLAC__stream_decoder_process_until_end_of_metadata (ogg_stream_decoder) ) {
            cleanup();
            return;
        }
    } else { // normal FLAC file/stream
        if ( pReader->can_seek() ) { // seekable
            seekable_stream_decoder = FLAC__seekable_stream_decoder_new();
            if ( !seekable_stream_decoder ) {
                return;
            }

            FLAC__seekable_stream_decoder_set_md5_checking ( seekable_stream_decoder, false );
            FLAC__seekable_stream_decoder_set_read_callback ( seekable_stream_decoder, readCallback_ );
            FLAC__seekable_stream_decoder_set_seek_callback ( seekable_stream_decoder, seekCallback_ );
            FLAC__seekable_stream_decoder_set_tell_callback ( seekable_stream_decoder, tellCallback_ );
            FLAC__seekable_stream_decoder_set_length_callback ( seekable_stream_decoder, lengthCallback_ );
            FLAC__seekable_stream_decoder_set_eof_callback ( seekable_stream_decoder, eofCallback_ );
            FLAC__seekable_stream_decoder_set_write_callback ( seekable_stream_decoder, writeCallback_ );
            FLAC__seekable_stream_decoder_set_metadata_callback ( seekable_stream_decoder, metadataCallback_ );
            FLAC__seekable_stream_decoder_set_error_callback ( seekable_stream_decoder, errorCallback_ );
            FLAC__seekable_stream_decoder_set_client_data ( seekable_stream_decoder, this );
            //FLAC__seekable_stream_decoder_set_metadata_respond ( seekable_stream_decoder, FLAC__METADATA_TYPE_VORBIS_COMMENT );

            if ( FLAC__seekable_stream_decoder_init (seekable_stream_decoder) != FLAC__SEEKABLE_STREAM_DECODER_OK ) {
                cleanup();
                return;
            }

            if ( !FLAC__seekable_stream_decoder_process_until_end_of_metadata (seekable_stream_decoder) ) {
                cleanup();
                return;
            }
        } else { // non-seekable
            stream_decoder = FLAC__stream_decoder_new();
            if ( !stream_decoder ) {
                return;
            }

            FLAC__stream_decoder_set_read_callback ( stream_decoder, readCallback_ns );
            FLAC__stream_decoder_set_write_callback ( stream_decoder, writeCallback_ns );
            FLAC__stream_decoder_set_metadata_callback ( stream_decoder, metadataCallback_ns );
            FLAC__stream_decoder_set_error_callback ( stream_decoder, errorCallback_ns );
            FLAC__stream_decoder_set_client_data ( stream_decoder, this );
            //FLAC__stream_decoder_set_metadata_respond ( stream_decoder, FLAC__METADATA_TYPE_VORBIS_COMMENT );

            if ( FLAC__stream_decoder_init (stream_decoder) != FLAC__STREAM_DECODER_SEARCH_FOR_METADATA ) {
                cleanup();
                return;
            }

            if ( !FLAC__stream_decoder_process_until_end_of_metadata (stream_decoder) ) {
                cleanup();
                return;
            }
        }
    }

    status_ok = true;
}

flac_decoder::~flac_decoder()
{
    cleanup();
    //if ( pReader ) delete pReader;
}

void flac_decoder::cleanup()
{
    if ( seekable_stream_decoder ) {
        if ( FLAC__seekable_stream_decoder_get_state (seekable_stream_decoder) != FLAC__SEEKABLE_STREAM_DECODER_UNINITIALIZED )
            FLAC__seekable_stream_decoder_finish ( seekable_stream_decoder );
        FLAC__seekable_stream_decoder_delete ( seekable_stream_decoder );
        seekable_stream_decoder = NULL;
    }

    if ( stream_decoder ) {
        if ( FLAC__stream_decoder_get_state (stream_decoder) != FLAC__STREAM_DECODER_UNINITIALIZED )
            FLAC__stream_decoder_finish ( stream_decoder );
        FLAC__stream_decoder_delete ( stream_decoder );
        stream_decoder = NULL;
    }

    if ( ogg_stream_decoder ) {
        if ( OggFLAC__stream_decoder_get_state (ogg_stream_decoder) != OggFLAC__STREAM_DECODER_UNINITIALIZED )
            OggFLAC__stream_decoder_finish ( ogg_stream_decoder );
        OggFLAC__stream_decoder_delete ( ogg_stream_decoder );
        ogg_stream_decoder = NULL;
    }

    if ( reservoir ) {
        free ( reservoir );
        reservoir = NULL;
        reservoir_size = 0;
        samples_in_reservoir = 0;
    }
}

static __int64 id3v2_tag_size ( Reader *r )
{
    unsigned char tmp[10];

    if ( r->read (tmp, sizeof (tmp)) != sizeof (tmp) ) {
        r->seek ( 0 );
        return -1;
    }

    // check id3-tag
    if ( memcmp (tmp, "ID3", 3) ) {
        r->seek ( 0 );
        return 0;
    }

    // read flags
    int Unsynchronisation = tmp[5] & 0x80;
    int ExtHeaderPresent  = tmp[5] & 0x40;
    int ExperimentalFlag  = tmp[5] & 0x20;
    int FooterPresent     = tmp[5] & 0x10;

    if ( tmp[5] & 0x0F ) {                              // not (yet???) allowed
        r->seek ( 0 );
        return 0;
    }

    if ( (tmp[6] | tmp[7] | tmp[8] | tmp[9]) & 0x80 ) { // not allowed
        r->seek ( 0 );
        return 0;
    }

    // read HeaderSize (syncsave: 4 * $0xxxxxxx = 28 significant bits)
    unsigned int ret;
    ret  = tmp[6] << 21;
    ret += tmp[7] << 14;
    ret += tmp[8] <<  7;
    ret += tmp[9];
    ret += 10;
    if ( FooterPresent ) ret += 10;

    r->seek ( ret );

    return ret;
}

int flac_decoder::is_flac ( Reader *r )
{
    if ( id3v2_tag_size (r) < 0 ) return 0;

    char head[4];
    int read = r->read ( head, 4 );
    r->seek ( 0 );

    if ( read < 4 || memcmp (head, "fLaC", 4) ) return 0;

    return 1;
}

int flac_decoder::is_ogg_flac ( Reader *r )
{
    char head[4];
    int read, read_total = 0, ret, eof = 0, state = -1;
    ogg_sync_state   oy;
    ogg_stream_state os;
    ogg_page         og;
    ogg_packet       op;

    read = r->read ( head, 4 );
    r->seek ( 0 );
    if ( read != 4 || memcmp (head, "OggS", 4) ) return 0;

    ogg_sync_init ( &oy );

    while ( !eof && read_total < 8192 ) {
        char *buf = ogg_sync_buffer ( &oy, 200 );
        if ( !buf ) break;

        read = r->read ( buf, 200 );
        read_total += read;
        if ( read < 200 ) eof = 1;
        if ( read <= 0 ) break;

        ogg_sync_wrote ( &oy, read );

        ret = ogg_sync_pageout ( &oy, &og );
        if ( ret < 0 ) break;
        if ( ret > 0 ) {
            state = 0;
            break;
        }
    }

    r->seek ( 0 );

    if ( state == 0 ) {
        ogg_stream_init ( &os, ogg_page_serialno (&og) );
        ogg_stream_pagein ( &os, &og );
        
        if ( ogg_stream_packetout (&os, &op) > 0 && op.bytes >= 4 ) {
            if ( !memcmp (op.packet, "fLaC", 4) ) state = 1;
        }

        ogg_stream_clear ( &os );
    }

    ogg_sync_clear ( &oy );

    return state == 1;
}

int flac_decoder::pack_pcm ( FLAC__byte *data, FLAC__int32 *input, unsigned wide_samples, unsigned channels, unsigned bps )
{
    FLAC__byte * const start = data;
    FLAC__int32 sample;
    unsigned samples = wide_samples * channels;
    const unsigned bytes_per_sample = bps / 8;

    while ( samples-- ) {
        sample = *input++;

        switch ( bps ) {
        case 8:
            data[0] = sample ^ 0x80;
            break;
        case 32:
            data[3] = (FLAC__byte)(sample >> 24);
        case 24:
            data[2] = (FLAC__byte)(sample >> 16);
        case 16:
            data[1] = (FLAC__byte)(sample >> 8);
            data[0] = (FLAC__byte)sample;
        }

        data += bytes_per_sample;
    }

    return data - start;
}

int flac_decoder::convert_to_float ( float *data, FLAC__int32 *input, unsigned wide_samples, unsigned channels, unsigned bps )
{
    unsigned samples = wide_samples * channels;
    const float mul = (float)(32768. / (1 << (bps-1)));

    while ( samples-- ) {
        *data++ = (float)((*input++) * mul);
    }

    return wide_samples * channels * sizeof(float);
}

// callbacks for seekable FLAC stream

FLAC__SeekableStreamDecoderReadStatus flac_decoder::readCallback_ ( const FLAC__SeekableStreamDecoder *decoder, FLAC__byte buffer[], unsigned *bytes, void *client_data )
{
    flac_decoder *instance = (flac_decoder *)client_data;
    *bytes = instance->pReader->read ( (char *)buffer, *bytes );
    return FLAC__SEEKABLE_STREAM_DECODER_READ_STATUS_OK;
}

FLAC__SeekableStreamDecoderSeekStatus flac_decoder::seekCallback_ ( const FLAC__SeekableStreamDecoder *decoder, FLAC__uint64 absolute_byte_offset, void *client_data )
{
    flac_decoder *instance = (flac_decoder *)client_data;
    if ( instance->pReader->seek (absolute_byte_offset) < 0 ) return FLAC__SEEKABLE_STREAM_DECODER_SEEK_STATUS_ERROR;
    return FLAC__SEEKABLE_STREAM_DECODER_SEEK_STATUS_OK;
}

FLAC__SeekableStreamDecoderTellStatus flac_decoder::tellCallback_ ( const FLAC__SeekableStreamDecoder *decoder, FLAC__uint64 *absolute_byte_offset, void *client_data )
{
    flac_decoder *instance = (flac_decoder *)client_data;
    __int64 pos = instance->pReader->get_position();
    if ( pos < 0 ) return FLAC__SEEKABLE_STREAM_DECODER_TELL_STATUS_ERROR;
    *absolute_byte_offset = pos;
    return FLAC__SEEKABLE_STREAM_DECODER_TELL_STATUS_OK;
}

FLAC__SeekableStreamDecoderLengthStatus flac_decoder::lengthCallback_ ( const FLAC__SeekableStreamDecoder *decoder, FLAC__uint64 *stream_length, void *client_data )
{
    flac_decoder *instance = (flac_decoder *)client_data;
    __int64 length = instance->pReader->get_length();
    if ( length < 0 ) return FLAC__SEEKABLE_STREAM_DECODER_LENGTH_STATUS_ERROR;
    *stream_length = length;
    return FLAC__SEEKABLE_STREAM_DECODER_LENGTH_STATUS_OK;
}

FLAC__bool flac_decoder::eofCallback_ ( const FLAC__SeekableStreamDecoder *decoder, void *client_data )
{
    flac_decoder *instance = (flac_decoder *)client_data;
    return instance->pReader->get_position() == instance->pReader->get_length();
}

FLAC__StreamDecoderWriteStatus flac_decoder::writeCallback_ ( const FLAC__SeekableStreamDecoder *decoder, const FLAC__Frame *frame, const FLAC__int32 * const buffer[], void *client_data )
{
    flac_decoder *instance = (flac_decoder *)client_data;
    const unsigned channels = instance->streaminfo.data.stream_info.channels, wide_samples = frame->header.blocksize;
    unsigned wide_sample, offset_sample, channel;

    if ( instance->abort_flag ) return FLAC__STREAM_DECODER_WRITE_STATUS_ABORT;

    int size = FLAC__MAX_BLOCK_SIZE * instance->streaminfo.data.stream_info.channels * 2;
    if ( size > instance->reservoir_size ) {
        void *backup = (void *)instance->reservoir;
        instance->reservoir = (FLAC__int32 *)realloc ( instance->reservoir, size * sizeof(FLAC__int32) );
        if ( !instance->reservoir ) {
            if ( backup ) free ( backup );
            return FLAC__STREAM_DECODER_WRITE_STATUS_ABORT;
        }
        instance->reservoir_size = size;
    }

    for ( offset_sample = instance->samples_in_reservoir * channels, wide_sample = 0; wide_sample < wide_samples; wide_sample++ )
        for ( channel = 0; channel < channels; channel++, offset_sample++ )
            instance->reservoir[offset_sample] = buffer[channel][wide_sample];

    instance->samples_in_reservoir += wide_samples;

    return FLAC__STREAM_DECODER_WRITE_STATUS_CONTINUE;
}

void flac_decoder::metadataCallback_ ( const FLAC__SeekableStreamDecoder *decoder, const FLAC__StreamMetadata *metadata, void *client_data )
{
    flac_decoder *instance = (flac_decoder *)client_data;

    if ( metadata->type == FLAC__METADATA_TYPE_STREAMINFO ) {
        instance->streaminfo = *metadata;
    }
}

void flac_decoder::errorCallback_ ( const FLAC__SeekableStreamDecoder *decoder, FLAC__StreamDecoderErrorStatus status, void *client_data )
{
    flac_decoder *instance = (flac_decoder *)client_data;
    (void)decoder;
    if ( status != FLAC__STREAM_DECODER_ERROR_STATUS_LOST_SYNC ) instance->abort_flag = true;
}

// callbacks for non-seekable FLAC stream

FLAC__StreamDecoderReadStatus flac_decoder::readCallback_ns ( const FLAC__StreamDecoder *decoder, FLAC__byte buffer[], unsigned *bytes, void *client_data )
{
    flac_decoder *instance = (flac_decoder *)client_data;
    *bytes = instance->pReader->read ( (char *)buffer, *bytes );
    return FLAC__STREAM_DECODER_READ_STATUS_CONTINUE;
    //FLAC__STREAM_DECODER_READ_STATUS_END_OF_STREAM
    //FLAC__STREAM_DECODER_READ_STATUS_ABORT
}

FLAC__StreamDecoderWriteStatus flac_decoder::writeCallback_ns ( const FLAC__StreamDecoder *decoder, const FLAC__Frame *frame, const FLAC__int32 * const buffer[], void *client_data )
{
    flac_decoder *instance = (flac_decoder *)client_data;
    const unsigned channels = instance->streaminfo.data.stream_info.channels, wide_samples = frame->header.blocksize;
    unsigned wide_sample, offset_sample, channel;

    if ( instance->abort_flag ) return FLAC__STREAM_DECODER_WRITE_STATUS_ABORT;

    int size = FLAC__MAX_BLOCK_SIZE * instance->streaminfo.data.stream_info.channels * 2;
    if ( size > instance->reservoir_size ) {
        void *backup = (void *)instance->reservoir;
        instance->reservoir = (FLAC__int32 *)realloc ( instance->reservoir, size * sizeof(FLAC__int32) );
        if ( !instance->reservoir ) {
            if ( backup ) free ( backup );
            return FLAC__STREAM_DECODER_WRITE_STATUS_ABORT;
        }
        instance->reservoir_size = size;
    }

    for ( offset_sample = instance->samples_in_reservoir * channels, wide_sample = 0; wide_sample < wide_samples; wide_sample++ )
        for ( channel = 0; channel < channels; channel++, offset_sample++ )
            instance->reservoir[offset_sample] = buffer[channel][wide_sample];

    instance->samples_in_reservoir += wide_samples;

    return FLAC__STREAM_DECODER_WRITE_STATUS_CONTINUE;
}

void flac_decoder::metadataCallback_ns ( const FLAC__StreamDecoder *decoder, const FLAC__StreamMetadata *metadata, void *client_data )
{
    flac_decoder *instance = (flac_decoder *)client_data;

    if ( metadata->type == FLAC__METADATA_TYPE_STREAMINFO ) {
        instance->streaminfo = *metadata;
    }
}

void flac_decoder::errorCallback_ns ( const FLAC__StreamDecoder *decoder, FLAC__StreamDecoderErrorStatus status, void *client_data )
{
    flac_decoder *instance = (flac_decoder *)client_data;
    (void)decoder;
    if ( status != FLAC__STREAM_DECODER_ERROR_STATUS_LOST_SYNC ) instance->abort_flag = true;
}

// callbacks for non-seekable OggFLAC stream

FLAC__StreamDecoderReadStatus flac_decoder::readCallback_ogg_ns ( const OggFLAC__StreamDecoder *decoder, FLAC__byte buffer[], unsigned *bytes, void *client_data )
{
    flac_decoder *instance = (flac_decoder *)client_data;
    *bytes = instance->pReader->read ( (char *)buffer, *bytes );
    return FLAC__STREAM_DECODER_READ_STATUS_CONTINUE;
    //FLAC__STREAM_DECODER_READ_STATUS_END_OF_STREAM
    //FLAC__STREAM_DECODER_READ_STATUS_ABORT
}

FLAC__StreamDecoderWriteStatus flac_decoder::writeCallback_ogg_ns ( const OggFLAC__StreamDecoder *decoder, const FLAC__Frame *frame, const FLAC__int32 * const buffer[], void *client_data )
{
    flac_decoder *instance = (flac_decoder *)client_data;
    const unsigned channels = instance->streaminfo.data.stream_info.channels, wide_samples = frame->header.blocksize;
    unsigned wide_sample, offset_sample, channel;

    if ( instance->abort_flag ) return FLAC__STREAM_DECODER_WRITE_STATUS_ABORT;

    int size = FLAC__MAX_BLOCK_SIZE * instance->streaminfo.data.stream_info.channels * 2;
    if ( size > instance->reservoir_size ) {
        void *backup = (void *)instance->reservoir;
        instance->reservoir = (FLAC__int32 *)realloc ( instance->reservoir, size * sizeof(FLAC__int32) );
        if ( !instance->reservoir ) {
            if ( backup ) free ( backup );
            return FLAC__STREAM_DECODER_WRITE_STATUS_ABORT;
        }
        instance->reservoir_size = size;
    }

    for ( offset_sample = instance->samples_in_reservoir * channels, wide_sample = 0; wide_sample < wide_samples; wide_sample++ )
        for ( channel = 0; channel < channels; channel++, offset_sample++ )
            instance->reservoir[offset_sample] = buffer[channel][wide_sample];

    instance->samples_in_reservoir += wide_samples;

    return FLAC__STREAM_DECODER_WRITE_STATUS_CONTINUE;
}

void flac_decoder::metadataCallback_ogg_ns ( const OggFLAC__StreamDecoder *decoder, const FLAC__StreamMetadata *metadata, void *client_data )
{
    flac_decoder *instance = (flac_decoder *)client_data;

    if ( metadata->type == FLAC__METADATA_TYPE_STREAMINFO ) {
        instance->streaminfo = *metadata;
    }
}

void flac_decoder::errorCallback_ogg_ns ( const OggFLAC__StreamDecoder *decoder, FLAC__StreamDecoderErrorStatus status, void *client_data )
{
    flac_decoder *instance = (flac_decoder *)client_data;
    (void)decoder;
    if ( status != FLAC__STREAM_DECODER_ERROR_STATUS_LOST_SYNC ) instance->abort_flag = true;
}

// -------------------------------------

int flac_encoder::compression_level ( flac_encoder_parameters &parameters, int level )
{
    if ( level < 0 || level > 8 ) return 0;

    parameters.channels = 0;
    parameters.bits_per_sample = 0;
    parameters.sample_rate = 0;
    parameters.total_samples = 0;
    parameters.create_ogg = false;
    parameters.verify = true;
    parameters.streamable_subset = true;

    switch ( level ) {
    case 0:
        parameters.do_exhaustive_model_search = false;
        parameters.do_escape_coding = false;
        parameters.do_mid_side_stereo = false;
        parameters.loose_mid_side_stereo = false;
        parameters.qlp_coeff_precision = 0;
        parameters.min_residual_partition_order = 2;
        parameters.max_residual_partition_order = 2;
        parameters.rice_parameter_search_dist = 0;
        parameters.max_lpc_order = 0;
        parameters.blocksize = 1152;
        break;

    case 1:
        parameters.do_exhaustive_model_search = false;
        parameters.do_escape_coding = false;
        parameters.do_mid_side_stereo = true;
        parameters.loose_mid_side_stereo = true;
        parameters.qlp_coeff_precision = 0;
        parameters.min_residual_partition_order = 2;
        parameters.max_residual_partition_order = 2;
        parameters.rice_parameter_search_dist = 0;
        parameters.max_lpc_order = 0;
        parameters.blocksize = 1152;
        break;

    case 2:
        parameters.do_exhaustive_model_search = false;
        parameters.do_escape_coding = false;
        parameters.do_mid_side_stereo = true;
        parameters.loose_mid_side_stereo = false;
        parameters.qlp_coeff_precision = 0;
        parameters.min_residual_partition_order = 0;
        parameters.max_residual_partition_order = 3;
        parameters.rice_parameter_search_dist = 0;
        parameters.max_lpc_order = 0;
        parameters.blocksize = 1152;
        break;

    case 3:
        parameters.do_exhaustive_model_search = false;
        parameters.do_escape_coding = false;
        parameters.do_mid_side_stereo = false;
        parameters.loose_mid_side_stereo = false;
        parameters.qlp_coeff_precision = 0;
        parameters.min_residual_partition_order = 3;
        parameters.max_residual_partition_order = 3;
        parameters.rice_parameter_search_dist = 0;
        parameters.max_lpc_order = 6;
        parameters.blocksize = 4608;
        break;

    case 4:
        parameters.do_exhaustive_model_search = false;
        parameters.do_escape_coding = false;
        parameters.do_mid_side_stereo = true;
        parameters.loose_mid_side_stereo = true;
        parameters.qlp_coeff_precision = 0;
        parameters.min_residual_partition_order = 3;
        parameters.max_residual_partition_order = 3;
        parameters.rice_parameter_search_dist = 0;
        parameters.max_lpc_order = 8;
        parameters.blocksize = 4608;
        break;

    case 5:
        parameters.do_exhaustive_model_search = false;
        parameters.do_escape_coding = false;
        parameters.do_mid_side_stereo = true;
        parameters.loose_mid_side_stereo = false;
        parameters.qlp_coeff_precision = 0;
        parameters.min_residual_partition_order = 3;
        parameters.max_residual_partition_order = 3;
        parameters.rice_parameter_search_dist = 0;
        parameters.max_lpc_order = 8;
        parameters.blocksize = 4608;
        break;

    case 6:
        parameters.do_exhaustive_model_search = false;
        parameters.do_escape_coding = false;
        parameters.do_mid_side_stereo = true;
        parameters.loose_mid_side_stereo = false;
        parameters.qlp_coeff_precision = 0;
        parameters.min_residual_partition_order = 0;
        parameters.max_residual_partition_order = 4;
        parameters.rice_parameter_search_dist = 0;
        parameters.max_lpc_order = 8;
        parameters.blocksize = 4608;
        break;

    case 7:
        parameters.do_exhaustive_model_search = true;
        parameters.do_escape_coding = false;
        parameters.do_mid_side_stereo = true;
        parameters.loose_mid_side_stereo = false;
        parameters.qlp_coeff_precision = 0;
        parameters.min_residual_partition_order = 0;
        parameters.max_residual_partition_order = 6;
        parameters.rice_parameter_search_dist = 0;
        parameters.max_lpc_order = 8;
        parameters.blocksize = 4608;
        break;

    case 8:
        parameters.do_exhaustive_model_search = true;
        parameters.do_escape_coding = false;
        parameters.do_mid_side_stereo = true;
        parameters.loose_mid_side_stereo = false;
        parameters.qlp_coeff_precision = 0;
        parameters.min_residual_partition_order = 0;
        parameters.max_residual_partition_order = 6;
        parameters.rice_parameter_search_dist = 0;
        parameters.max_lpc_order = 12;
        parameters.blocksize = 4608;
        break;
    }

    return 1;
}

flac_encoder::flac_encoder ( Reader *r, flac_encoder_parameters &parameters, FLAC__StreamMetadata *comments=0 )
{
    status_ok = false;

    seekable_stream_encoder = NULL;
    ogg_stream_encoder = NULL;
    metadata = NULL;
    metadata_blocks = 0;
    error_message[0] = '\0';

    if ( parameters.channels == 0 || parameters.bits_per_sample == 0 || parameters.sample_rate == 0 ) return;

    channels = parameters.channels;
    bits_per_sample = parameters.bits_per_sample;

    pReader = r;

    encode_buffer = (FLAC__int32 *)malloc ( FLAC__MAX_BLOCK_SIZE * channels * 4 * sizeof(FLAC__int32) );
    if ( !encode_buffer ) {
        strcpy ( error_message, "Memory allocation failed." );
        return;
    }

    int meta_count = 1; // padding
    if ( comments ) meta_count++;

    metadata = (FLAC__StreamMetadata **)malloc ( sizeof (FLAC__StreamMetadata *) * meta_count );
    if ( !metadata ) {
        cleanup();
        strcpy ( error_message, "Memory allocation failed." );
        return;
    }

    if ( comments ) {
        metadata[metadata_blocks] = comments;
        metadata_blocks++;
    }

    metadata[metadata_blocks] = FLAC__metadata_object_new ( FLAC__METADATA_TYPE_PADDING );
    if ( !metadata[metadata_blocks] ) {
        cleanup();
        strcpy ( error_message, "Memory allocation failed." );
        return;
    }

    metadata[metadata_blocks]->length = 4096;
    metadata_blocks++;

    metadata[metadata_blocks] = FLAC__metadata_object_new ( FLAC__METADATA_TYPE_PADDING );
    if ( !metadata[metadata_blocks] ) {
        cleanup();
        return;
    }

    metadata[metadata_blocks]->length = 4096;
    metadata_blocks++;

    if ( parameters.channels != 2 ) {
        parameters.do_mid_side_stereo = false;
        parameters.loose_mid_side_stereo = false;
    }

    if ( !parameters.create_ogg ) { // encode normal FLAC file
        seekable_stream_encoder = FLAC__seekable_stream_encoder_new();
        if ( !seekable_stream_encoder ) {
            cleanup();
            strcpy ( error_message, "Failed to create new encoder." );
            return;
        }

        FLAC__seekable_stream_encoder_set_streamable_subset ( seekable_stream_encoder, parameters.streamable_subset );
        FLAC__seekable_stream_encoder_set_channels ( seekable_stream_encoder, parameters.channels );
        FLAC__seekable_stream_encoder_set_bits_per_sample ( seekable_stream_encoder, parameters.bits_per_sample );
        FLAC__seekable_stream_encoder_set_sample_rate ( seekable_stream_encoder, parameters.sample_rate );

        FLAC__seekable_stream_encoder_set_do_exhaustive_model_search ( seekable_stream_encoder, parameters.do_exhaustive_model_search );
        FLAC__seekable_stream_encoder_set_do_escape_coding ( seekable_stream_encoder, parameters.do_escape_coding );
        FLAC__seekable_stream_encoder_set_do_mid_side_stereo ( seekable_stream_encoder, parameters.do_mid_side_stereo );
        FLAC__seekable_stream_encoder_set_loose_mid_side_stereo ( seekable_stream_encoder, parameters.loose_mid_side_stereo );
        FLAC__seekable_stream_encoder_set_qlp_coeff_precision ( seekable_stream_encoder, parameters.qlp_coeff_precision );
        FLAC__seekable_stream_encoder_set_min_residual_partition_order ( seekable_stream_encoder, parameters.min_residual_partition_order );
        FLAC__seekable_stream_encoder_set_max_residual_partition_order ( seekable_stream_encoder, parameters.max_residual_partition_order );
        FLAC__seekable_stream_encoder_set_rice_parameter_search_dist ( seekable_stream_encoder, parameters.rice_parameter_search_dist );
        FLAC__seekable_stream_encoder_set_max_lpc_order ( seekable_stream_encoder, parameters.max_lpc_order );
        FLAC__seekable_stream_encoder_set_blocksize ( seekable_stream_encoder, parameters.blocksize );

        FLAC__seekable_stream_encoder_set_total_samples_estimate ( seekable_stream_encoder, parameters.total_samples );
        FLAC__seekable_stream_encoder_set_verify ( seekable_stream_encoder, parameters.verify );

        FLAC__seekable_stream_encoder_set_write_callback ( seekable_stream_encoder, writeCallback_ );
        FLAC__seekable_stream_encoder_set_seek_callback ( seekable_stream_encoder, seekCallback_ );
        FLAC__seekable_stream_encoder_set_client_data ( seekable_stream_encoder, this );

        FLAC__seekable_stream_encoder_set_metadata ( seekable_stream_encoder, metadata, metadata_blocks );

        if ( FLAC__seekable_stream_encoder_init (seekable_stream_encoder) != FLAC__SEEKABLE_STREAM_ENCODER_OK ) {
            get_state();
            cleanup();
            return;
        }
    } else { // encode OggFLAC file
        ogg_stream_encoder = OggFLAC__stream_encoder_new();
        if ( !ogg_stream_encoder ) {
            cleanup();
            strcpy ( error_message, "Failed to create new encoder." );
            return;
        }

        OggFLAC__stream_encoder_set_streamable_subset ( ogg_stream_encoder, parameters.streamable_subset );
        OggFLAC__stream_encoder_set_channels ( ogg_stream_encoder, parameters.channels );
        OggFLAC__stream_encoder_set_bits_per_sample ( ogg_stream_encoder, parameters.bits_per_sample );
        OggFLAC__stream_encoder_set_sample_rate ( ogg_stream_encoder, parameters.sample_rate );

        OggFLAC__stream_encoder_set_do_exhaustive_model_search ( ogg_stream_encoder, parameters.do_exhaustive_model_search );
        OggFLAC__stream_encoder_set_do_escape_coding ( ogg_stream_encoder, parameters.do_escape_coding );
        OggFLAC__stream_encoder_set_do_mid_side_stereo ( ogg_stream_encoder, parameters.do_mid_side_stereo );
        OggFLAC__stream_encoder_set_loose_mid_side_stereo ( ogg_stream_encoder, parameters.loose_mid_side_stereo );
        OggFLAC__stream_encoder_set_qlp_coeff_precision ( ogg_stream_encoder, parameters.qlp_coeff_precision );
        OggFLAC__stream_encoder_set_min_residual_partition_order ( ogg_stream_encoder, parameters.min_residual_partition_order );
        OggFLAC__stream_encoder_set_max_residual_partition_order ( ogg_stream_encoder, parameters.max_residual_partition_order );
        OggFLAC__stream_encoder_set_rice_parameter_search_dist ( ogg_stream_encoder, parameters.rice_parameter_search_dist );
        OggFLAC__stream_encoder_set_max_lpc_order ( ogg_stream_encoder, parameters.max_lpc_order );
        OggFLAC__stream_encoder_set_blocksize ( ogg_stream_encoder, parameters.blocksize );

        OggFLAC__stream_encoder_set_total_samples_estimate ( ogg_stream_encoder, parameters.total_samples );
        OggFLAC__stream_encoder_set_verify ( ogg_stream_encoder, parameters.verify );

        OggFLAC__stream_encoder_set_write_callback ( ogg_stream_encoder, writeCallback_ogg_ns );
        OggFLAC__stream_encoder_set_client_data ( ogg_stream_encoder, this );

        OggFLAC__stream_encoder_set_metadata ( ogg_stream_encoder, metadata, metadata_blocks );

        if ( OggFLAC__stream_encoder_init (ogg_stream_encoder) != OggFLAC__STREAM_ENCODER_OK ) {
            get_state();
            cleanup();
            return;
        }
    }

    status_ok = true;
}

flac_encoder::~flac_encoder()
{
    cleanup();
    //if ( pReader ) delete pReader;
}

int flac_encoder::get_state()
{
    const char *p = NULL;

    if ( seekable_stream_encoder ) {
        p = FLAC__seekable_stream_encoder_get_resolved_state_string ( seekable_stream_encoder );
    }
    else if ( ogg_stream_encoder ) {
        p = OggFLAC__StreamEncoderStateString[(unsigned)OggFLAC__stream_encoder_get_state (ogg_stream_encoder)];
    }

    if ( p ) strcpy ( error_message, p );

    return 1;
}

int flac_encoder::encode ( const void *buffer, unsigned int samples, int source_bps, double float_scale )
{
    unsigned int total = samples;
    FLAC__byte *buf = (FLAC__byte *)buffer;

    while ( samples > 0 ) {
        unsigned int _s = (samples > FLAC__MAX_BLOCK_SIZE) ? FLAC__MAX_BLOCK_SIZE : samples;

        if ( source_bps == 32 && bits_per_sample < 32 ) {
            FLAC_float32 ( (float *)buf, encode_buffer, _s, float_scale );
        }
        else if ( source_bps == 64 && bits_per_sample < 32 ) {
            FLAC_float64 ( (double *)buf, encode_buffer, _s, float_scale );
        }
        else {
            FLAC_pcm ( buf, encode_buffer, _s );
        }

        if ( seekable_stream_encoder ) {
            if ( !FLAC__seekable_stream_encoder_process_interleaved (seekable_stream_encoder, encode_buffer, _s) ) {
                get_state();
                return 0;
            }
        }
        else if ( ogg_stream_encoder ) {
            if ( !OggFLAC__stream_encoder_process_interleaved (ogg_stream_encoder, encode_buffer, _s) ) {
                get_state();
                return 0;
            }
        }
        else return 0;

        samples -= _s;
        buf += _s * channels * ((source_bps > 0) ? source_bps : bits_per_sample)/8;
    }

    return total;
}

void flac_encoder::cleanup()
{
    if ( seekable_stream_encoder ) {
        if ( FLAC__seekable_stream_encoder_get_state (seekable_stream_encoder) != FLAC__SEEKABLE_STREAM_ENCODER_UNINITIALIZED )
            FLAC__seekable_stream_encoder_finish ( seekable_stream_encoder );
        FLAC__seekable_stream_encoder_delete ( seekable_stream_encoder );
        seekable_stream_encoder = NULL;
    }

    if ( ogg_stream_encoder ) {
        if ( OggFLAC__stream_encoder_get_state (ogg_stream_encoder) != OggFLAC__STREAM_ENCODER_UNINITIALIZED )
            OggFLAC__stream_encoder_finish ( ogg_stream_encoder );
        OggFLAC__stream_encoder_delete ( ogg_stream_encoder );
        ogg_stream_encoder = NULL;
    }

    if ( metadata ) {
        for ( int i = 0; i < metadata_blocks; i++ ) {
            FLAC__metadata_object_delete ( metadata[i] );
        }
        free ( metadata );
        metadata = NULL;
        metadata_blocks = 0;
    }

    if ( encode_buffer ) {
        free ( encode_buffer );
        encode_buffer = NULL;
    }
}

int flac_encoder::FLAC_pcm ( const FLAC__byte *data, FLAC__int32 *output, unsigned wide_samples )
{
    unsigned samples = wide_samples * channels;

    switch ( bits_per_sample ) {
    case 8:
        {
            unsigned char *src = (unsigned char *)data;
            while ( samples-- ) {
                *output++ = (int)(*src++) - 128;
            }
        }
        break;

    case 16:
        {
            short *src = (short *)data;
            while ( samples-- ) {
                *output++ = *src++;
            }
        }
        break;

    case 24:
        {
            unsigned char *src = (unsigned char *)data;
            while ( samples-- ) {
                int sample = (((int)src[0] <<  0) |
                              ((int)src[1] <<  8) |
                              ((int)src[2] << 16));

                if ( sample & (1<<23) ) {
                    sample = - (((~sample) & 0x00FFFFFF)+1);
                }

                *output++ = (FLAC__int32)sample;
                src += 3;
            }
        }
        break;

    case 32:
        {
            int *src = (int *)data;
            while ( samples-- ) {
                *output++ = *src++;
            }
        }
        break;

    default:
        return 0;
    }

    return wide_samples;
}

int flac_encoder::FLAC_float32 ( const float *data, FLAC__int32 *output, unsigned wide_samples, double scale )
{
    unsigned samples = wide_samples * channels;
    const float mul = (float)((1 << (bits_per_sample-1)) / scale);
    int min = -(1 << (bits_per_sample-1));
    int max = +(1 << (bits_per_sample-1)) - 1;

    while ( samples-- ) {
        int sample = (int)((*data++) * mul);
        if ( sample < min ) sample = min;
        if ( sample > max ) sample = max;
        *output++ = (FLAC__int32)sample;
    }

    return wide_samples;
}

int flac_encoder::FLAC_float64 ( const double *data, FLAC__int32 *output, unsigned wide_samples, double scale )
{
    unsigned samples = wide_samples * channels;
    const double mul = (double)((1 << (bits_per_sample-1)) / scale);
    int min = -(1 << (bits_per_sample-1));
    int max = +(1 << (bits_per_sample-1)) - 1;

    while ( samples-- ) {
        int sample = (int)((*data++) * mul);
        if ( sample < min ) sample = min;
        if ( sample > max ) sample = max;
        *output++ = (FLAC__int32)sample;
    }

    return wide_samples;
}

// callbacks for seekable FLAC stream

FLAC__StreamEncoderWriteStatus flac_encoder::writeCallback_ ( const FLAC__SeekableStreamEncoder *encoder, const FLAC__byte buffer[], unsigned bytes, unsigned samples, unsigned current_frame, void *client_data )
{
    flac_encoder *instance = (flac_encoder *)client_data;
    if ( bytes == 0 ) return FLAC__STREAM_ENCODER_WRITE_STATUS_OK;
    if ( instance->pReader->write ((char *)buffer, (int)bytes) != (int)bytes ) return FLAC__STREAM_ENCODER_WRITE_STATUS_FATAL_ERROR;
    return FLAC__STREAM_ENCODER_WRITE_STATUS_OK;
}

FLAC__SeekableStreamEncoderSeekStatus flac_encoder::seekCallback_ ( const FLAC__SeekableStreamEncoder *encoder, FLAC__uint64 absolute_byte_offset, void *client_data )
{
    flac_encoder *instance = (flac_encoder *)client_data;
    if ( instance->pReader->seek (absolute_byte_offset) != (__int64)absolute_byte_offset ) return FLAC__SEEKABLE_STREAM_ENCODER_SEEK_STATUS_ERROR;
    return FLAC__SEEKABLE_STREAM_ENCODER_SEEK_STATUS_OK;
}

// callbacks for non-seekable OggFLAC stream

FLAC__StreamEncoderWriteStatus flac_encoder::writeCallback_ogg_ns ( const OggFLAC__StreamEncoder *encoder, const FLAC__byte buffer[], unsigned bytes, unsigned samples, unsigned current_frame, void *client_data )
{
    flac_encoder *instance = (flac_encoder *)client_data;
    if ( bytes == 0 ) return FLAC__STREAM_ENCODER_WRITE_STATUS_OK;
    if ( instance->pReader->write ((void *)buffer, (int)bytes) != (int)bytes ) return FLAC__STREAM_ENCODER_WRITE_STATUS_FATAL_ERROR;
    return FLAC__STREAM_ENCODER_WRITE_STATUS_OK;
}
